%% A k-space filtering technique for vibration sonoelastography
%
% Master thesis for Mikkel Henninen June 2016
%
% authors: Alfonso Rodriguez-Molares <alfonso.r.molares@ntnu.no>
%          Mikkel Henninen <mikkeloy@stud.ntnu.no>

clear all;
close all;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% File loop
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for filenumber = 10:10 %3.5, 2.5, 1, 0.5, hard,soft

    tic
    close all
    clearvars -except filenumber
    saving = 1;
    add_noise=0;
    use_fitting=1;     

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Loading/downloading data
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    url='http://folk.ntnu.no/alfonsom/elastography_data/Experiment_02_05_2016/';

    Change here for you local path
    folder='Data\numerical\';
    savefolder = 'C:\Users\Min\Documents\MATLAB\Master\Figs\Final results\Experimental_alfonso\';
    

    switch (filenumber)
        case 0
            % Hard Cyst 500 Hz, 10000PFR, take 1
            filename='Harmonic_PVA_square_hard_cyst_take01_500_10000.hdf5';
            acoustic_frequency=500;    % excitation frequency [Hz]
            save_name = 'exp_hard_cyst_500Hz_take1_';
            soft = false;
        case 1
            % Hard Cyst 500 Hz, 10000PFR, take 2
            filename='Harmonic_PVA_square_hard_cyst_take02_500_10000.hdf5';
            acoustic_frequency=500;    % excitation frequency [Hz]
            save_name = 'exp_hard_cyst_500Hz_take2_';
            soft = false;
        case 2
            % Hard Cyst 1000 Hz, 10000PFR, take 1
            filename='Harmonic_PVA_square_hard_cyst_take03_1000_10000.hdf5';
            acoustic_frequency=1000;    % excitation frequency [Hz]
            save_name = 'exp_hard_cyst_1000Hz_take1_';
            soft = false;
        case 3
            % Hard Cyst 1000 Hz, 10000PFR, take 2
            filename='Harmonic_PVA_square_hard_cyst_take04_1000_10000.hdf5';
            acoustic_frequency=1000;    % excitation frequency [Hz]
            save_name = 'exp_hard_cyst_1000Hz_take2_';
            soft = false;
        case 4
            % Hard Cyst 300 Hz, 10000PFR, take 1
            filename='Harmonic_PVA_square_hard_cyst_take05_300_10000.hdf5';
            acoustic_frequency=300;    % excitation frequency [Hz]
            save_name = 'exp_hard_cyst_300Hz_take1_';
            soft = false;
        case 5
            % Hard Cyst 300 Hz, 10000PFR, take 2
            filename='Harmonic_PVA_square_hard_cyst_take06_300_10000.hdf5';
            acoustic_frequency=300;    % excitation frequency [Hz]
            save_name = 'exp_hard_cyst_300Hz_take2_';
            soft = false;
        case 6
            % Soft Cyst 500 Hz, 10000PFR, take 1
            filename='Harmonic_PVA_square_soft_cyst_take01_500_10000.hdf5';
            acoustic_frequency=500;    % excitation frequency [Hz]
            save_name = 'exp_soft_cyst_500Hz_take1_';
            soft = true;
        case 7
            % soft Cyst 500 Hz, 10000PFR, take 2
            filename='Harmonic_PVA_square_soft_cyst_take02_500_10000.hdf5';
            acoustic_frequency=500;    % excitation frequency [Hz]
            save_name = 'exp_soft_cyst_500Hz_take2_';
            soft = true;
        case 8
            % soft Cyst 1000 Hz, 10000PFR, take 1
            filename='Harmonic_PVA_square_soft_cyst_take05_1000_10000.hdf5';
            acoustic_frequency=1000;    % excitation frequency [Hz]
            save_name = 'exp_soft_cyst_1000Hz_take1_';
            soft = true;
        case 9
            % soft Cyst 1000 Hz, 10000PFR, take 2
            filename='Harmonic_PVA_square_soft_cyst_take06_1000_10000.hdf5';
            acoustic_frequency=1000;    % excitation frequency [Hz]
            save_name = 'exp_soft_cyst_1000Hz_take2_';
            soft = true;
        case 10
            % soft Cyst 300 Hz, 10000PFR, take 1
            filename='Harmonic_PVA_square_soft_cyst_take03_300_10000.hdf5';
            acoustic_frequency=300;    % excitation frequency [Hz]
            save_name = 'exp_soft_cyst_300Hz_take1_';
            soft = true;
        case 11
            % soft Cyst 300 Hz, 10000PFR, take 2
            filename='Harmonic_PVA_square_soft_cyst_take04_300_10000.hdf5';
            acoustic_frequency=300;    % excitation frequency [Hz]
            save_name = 'exp_soft_cyst_300Hz_take2_';
            soft = true;
    end
    
    % check if the case has been calculated
    procesdatafilename=[save_name 'workspace_after_fitting_and_kspace.mat'];
    if numel(dir(procesdatafilename))
        load(procesdatafilename);
    else

    %% Loading/downloading data
    % we check if the file is on disk; if not we download it from the webpage
    path='C:\Users\Min\Documents\MATLAB\Master\data\Experiment_02_05_2016\';
    a=dir([path filename]);
    if not(numel(a))
        disp('Downloading data...');
        urlwrite([url filename],[path filename]);
    end
    hf=huff([path filename]);
    hf.read();
    cpw_dataset=hf.ultrasound_signal{1};
    
    %% PRF and Nyquish
    PRF = 3450;
    dt=1/PRF;
    vNyq = cpw_dataset.c0*PRF/(4*cpw_dataset.modulation_frequency);
    
    %% range of velocities of interest
    NFFT=1024;                              % size of the fft
    cs_min=1;                               % minimum shear wave speed we will visualize [m/s]
    cs_max=20;                              % minimum shear wave speed we will visualize [m/s]
    lambda_max=cs_max/acoustic_frequency;
    lambda_min=cs_min/acoustic_frequency;
    
    %% initial spatial sampling
    oversampling=2;                         % oversampling to eliminate noise
    Fs_over=2*oversampling/lambda_min;      % sampling spatial frequency [Hz]
    dx=1/Fs_over;                           % spatial step [m]
    Fs = 1/dx;
    ds = 1/Fs;
    n_frames = 99;
    drawFigs = 0;
    
    %% beamforming
    recons=reconstruction();                    % create a reconstruction object
    recons.scan=linear_scan();                  % define the scan
    x_axis=(-19e-3:dx:19e-3).';                 % x vector [m]  -> this vectors were also wrong
    z_axis=(2e-3:dx:35e-3).';                   % z vector [m]
    recons.scan.x_axis=x_axis;
    recons.scan.z_axis=z_axis;
    recons.orientation=orientation();
    recons.orientation.transmit_beam=beam(1,E.apodization_type.none);
    recons.orientation.receive_beam=beam(1.2,E.apodization_type.hanning);
    cpw_dataset.image_reconstruction(recons,E.implementation.simple_matlab);  % request reconstruction
    %recons.show();                             % show
    IQ=squeeze(recons.data);                   % reconstructed IQ data
    
%     for i = 1:100
%         pcolor(x_axis,z_axis,abs(recons.data(:,:,1,i))); shading flat; colormap gray; colorbar; hold on;
%         pause(0.1)
%     end
    
    %% Tissue Doppler Imaging
    R1 = conj(IQ(:,:,1:end-1)).*IQ(:,:,2:end);  % correlation
    t=(1:size(R1,3))*dt; t=t-t(1);              % time vector
    Ns = 5;
    sAvg = ones(Ns)./Ns^2;                      % spatial average
    for k=1:size(R1,3)
        R1(:,:,k) = filter2(sAvg,R1(:,:,k));
    end
    TDI = angle(R1)/pi*vNyq; % m/s
    %TDI = bsxfun(@minus,TDI,mean(mean(TDI,1),2));  % remove mean value
    
    [Nz Nx Nf]=size(TDI);
    
        figure;
        colormap jet;
        filename = sprintf('TDI_Hard_300.gif');
        for n=1:Nf
            imagesc(recons.scan.x_axis*1e3,recons.scan.z_axis*1e3,TDI(:,:,n)); axis equal tight;
            caxis([-max(abs(TDI(:))) max(abs(TDI(:)))]);
            colorbar;
            title(sprintf('Original, frame %d',n));
            %colormap(gray);
            drawnow;
            pause(0.1)
            
    frame = getframe(gcf);
    im = frame2im(frame);
    [imind,cm] = rgb2ind(im,256);
    if(n==1)
        imwrite(imind,cm,filename,'gif', 'Loopcount',inf);
    else
        imwrite(imind,cm,filename,'gif','WriteMode','append','DelayTime', 0.1);
    end
        end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Manual detection jumps in the sequence
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    sig=squeeze(TDI(1,1,:));
    figure;
    plot(sig); axis tight; hold on;
    title('Mark any discontinuities in the sequence. Enter when done');
    notempty=1;
    Ndis=[];
    while (notempty)
        x1=ginput(1);
        if ~isempty(x1)
            dis=round(x1(1));
            plot([dis dis],[min(sig) max(sig)],'r--'); hold on;
            Ndis=[Ndis dis];
        else
            notempty=0;
        end
    end
    hold off;
    
    % we try to identify which sequence we should use
    Ndis=sort(Ndis);
    seqs=[];
    seqs(1,:)=[1 Ndis(1) Ndis(1)];
    for n=1:length(Ndis)-1;
        seqs(n+1,:)=[Ndis(n) Ndis(n+1) Ndis(n+1)-Ndis(n)+1];
    end
    seqs(length(Ndis)+1,:)=[Ndis(end) size(R1,3) size(R1,3)-Ndis(end)+1];
    [X I]=max(seqs(:,3));
    N_ini=seqs(I,1);
    N_end=seqs(I,2);
    
    FTDI = TDI;
    drawFigs = 0;
    freq_iter_num = 0;
    freq_iter = freq_iter_num;
    speeds = [0 0];                                 %Average speed outside and inside the cyst
    cs = zeros(1,n_frames*(freq_iter_num+1));             %
    cs_2 = zeros(1,n_frames*(freq_iter_num+1));
    cscm = zeros(1,n_frames*(freq_iter_num+1));
    low_speed_filtered_space_cum = zeros(length(z_axis),length(x_axis));  % for summing up the cumulative fields
    NFFT = 1024;
    high_speed_filtered_space_cum = zeros(length(z_axis),length(x_axis));  % for summing up the cumulative fields
    

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %% k-space transform
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        % MIKEL: You were using an old version of the k-space transform. Not
        % that there was something wrong with it, but I think I have made some
        % small improvements in the function "kspace" below.

        %% compute the k-space transform
        disp('Computing k-space. This may take a while');
        [kfield,kz,kx]=kspace(FTDI,ds,0.25,4,false);  
        
        % showing kspace
        cs1=5;
        cs2=10;
        theta=linspace(0,2*pi,180);
        ks1=2*pi*acoustic_frequency/cs1;
        ks2=2*pi*acoustic_frequency/cs2;

        figure;
        pcolor(kx*1e-3,kz*1e-3,sum(abs(kfield),3)); axis equal tight; hold on; shading flat;
        xlabel('k_x [rad/mm]');ylabel('k_z [rad/mm]');
        plot(ks1*cos(theta)*1e-3,ks1*sin(theta)*1e-3,'y--'); 
        plot(ks2*cos(theta)*1e-3,ks2*sin(theta)*1e-3,'r--'); 
        xlim([-2*ks1 2*ks1]/1e3);
        ylim([-2*ks1 2*ks1]/1e3);
        title('k-space');
        legend('',sprintf('cs=%0.1f m/s',cs1),sprintf('cs=%0.1f m/s',cs2));
        drawnow;

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %% spectrum estimation
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        
        alpha=linspace(0,2*pi,180);
        c_axis=linspace(cs_max,1,200);
        k_axis=2*pi*acoustic_frequency./c_axis;
        kx_axis=k_axis.'*sin(alpha); kz_axis=k_axis.'*cos(alpha);

        spectrum=zeros(size(c_axis)).';
        for n=1:size(kfield,3)
            kf=interp2(kx,kz,kfield(:,:,n),kx_axis,kz_axis,'linear');
            spectrum=spectrum+(sum(abs(kf),2)); 
        end

%         figure;
%         plot(k_axis,spectrum/max(spectrum),'k-'); hold on; grid on;
%         plot([2*pi*acoustic_frequency/cs2 2*pi*acoustic_frequency/cs2],[0 1],'b--');
%         plot([2*pi*acoustic_frequency/cs1 2*pi*acoustic_frequency/cs1],[0 1],'r--');
%         xlabel('k [rad/m]');
%         legend('cyst + background','cyst','background');
%         title(sprintf('k-space spectrum %0.0f Hz',acoustic_frequency));

        % remove NaNs
        spectrum(isnan(spectrum))=0;
    
        % save workspace for future re-runs
        save(procesdatafilename);
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% feature extraction
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        datafilename=[save_name];
                
        cs_vector=[];
        filter=[];
            %% peak detection

            % first we smothen the spectrum a bit to remove noisy peaks
            smoothing_window_size=15;
            f_spectrum=filtfilt(hanning(smoothing_window_size)/sum(hanning(smoothing_window_size)),1,spectrum);

            % we detect the peaks
            [pks,locs,widths,proms] = findpeaks(f_spectrum,k_axis);

            % we sort them by prominence (i.e. importance)
            [dummy indexes]=sort(proms,'descend');
            pks=pks(indexes);
            locs=locs(indexes);
            widths=widths(indexes);
            proms=proms(indexes);
            cs_vector=locs;

            % design ring filter automatically

            
            for m=1:length(locs)
                filter(m,:)=(k_axis>locs(m)-widths(m))&(k_axis<locs(m)+widths(m));
            end

%                 % show spectrum, peaks & filters
%                 figure;
%                 plot(c_axis,spectrum./max(spectrum),'k-'); hold on; grid on;
%                 plot(c_axis,f_spectrum./max(spectrum),'b--'); 
%                 plot(2*pi*acoustic_frequency./locs,pks./max(spectrum),'r^'); 
%                 for m=1:length(locs)
%                     plot(2*pi*acoustic_frequency./[locs(m)-widths(m)/2 locs(m)+widths(m)/2],[pks(m) pks(m)]./max(spectrum),'r-s'); 
%                     plot(c_axis,filter(m,:),'--');
%                 end
%                 xlabel('Shear wave speed [m/s]');
%                 ylabel('Normalized energy');


        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %% Filtering
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        components=2;%size(filter,1)
        [KX KZ]=meshgrid(kx,kz);
        K=sqrt(KX.^2+KZ.^2);
        cs=zeros(1,1,components);
        map=zeros(size(TDI,1),size(TDI,2),components);
        for m=1:components
            % filtering
            kfiltered=bsxfun(@times,kfield,interp1(k_axis,filter(m,:),K,'linear',0));

            % inverse fft
            disp('Computing inverse FFT. This may take a while');
            ffield=ifft2(ifftshift(kfiltered));
            ffield=real(ifft2(ifftshift(kfiltered)));
            ffield=ffield(1:Nz,1:Nx,:,:);

            % storing map and
            cs(1,1,m)=2*pi*acoustic_frequency/cs_vector(m);
            map(:,:,m)=sum(abs(ffield),3);

%             % showing space
%             figure(106);
%             subplot(1,4,1); hold off; 
%             imagesc(x_axis,z_axis,ctdi); axis equal tight; hold on; 
%             title('Space');
% 
%             % showing kspace
%             subplot(1,4,2); hold off;
%             imagesc(kx*1e-3,kz*1e-3,abs(kfield)); axis equal tight; hold on; shading flat;
%             xlabel('k_x [rad/mm]');ylabel('k_z [rad/mm]');
%             xlim([-ks1 ks1]*1.5*1e-3);
%             ylim([-ks1 ks1]*1.5*1e-3);
%             title('k-space');
% 
%             % showing filtered kspace
%             subplot(1,4,3); hold off;
%             imagesc(kx*1e-3,kz*1e-3,abs(kfiltered)); axis equal tight; hold on; shading flat;
%             xlabel('k_x [rad/mm]');ylabel('k_z [rad/mm]');
%             xlim([-ks1 ks1]*1.5*1e-3);
%             ylim([-ks1 ks1]*1.5*1e-3);
%             title('filtered k-space');   
% 
%             % showing space
%             subplot(1,4,4); hold off; 
%             imagesc(x_axis,z_axis,ffield); axis equal tight; hold on; 
% %             title('Space');
% 
%             figure(101);
%             imagesc(x_axis*1e3,z_axis*1e3,map(:,:,m)); axis equal tight; hold on; 
%             %plot(cx*1e3+cyst_radius*1e3*sin(linspace(0,2*pi,180)),cz*1e3+cyst_radius*1e3*cos(linspace(0,2*pi,180)),'r--')
%             xlabel('x [mm]');
%             ylabel('z [mm]');
%             title(sprintf('Filtered region for %0.1f m/s',cs(m)));
%             if saving
%             saveas(gcf,[savefolder sprintf('%smethod_%d_filter_%d',save_name,extraction_method,m)], 'jpg');
%             end
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%  Global map
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % computing global map as a weigthed mean
        csm=sum(bsxfun(@times,map,cs),3)./sum(map,3);
    
        % smoothing a bit with a media filter        
        csmm=medfilt2(csm,[5 5]);

        figure(102);
        imagesc(x_axis*1e3,z_axis*1e3,csmm); axis equal tight; shading flat; colormap jet; hold on;
        %plot(cx*1e3+cyst_radius*1e3*sin(linspace(0,2*pi,180)),cz*1e3+cyst_radius*1e3*cos(linspace(0,2*pi,180)),'k--')
        caxis([0 10]);
        c=colorbar; xlabel(c,'c_s [m/s]');
        xlabel('x [mm]');
        ylabel('z [mm]');
        title('Shear wave velocity map [m/s]');
        if saving
        saveas(gcf,[savefolder sprintf('%smethod_%d_elasticity_map',save_name,extraction_method)], 'jpg');
        end
    end          
end

  